import click
from .oai import (
    # ALL_TOOLS_AVAILABLE,
    create_assistant_wrapper,
    get_assistants,
    list_threads,
    create_thread,
    list_files,
    upload_file,
    save_local_message,
    create_run,
    wait_for_or_cancel_run,
    get_messages,
    _get_assistant_path,
    load_instructions,
    save_instructions,
    client,
)
from . import FilePathType
import textwrap
from datetime import datetime


def wrap_text(text, width=150, dash=False):
    wrapper = textwrap.TextWrapper(
        width=width, break_long_words=False, replace_whitespace=False
    )
    wrapped_lines = wrapper.wrap(text)

    if dash:
        wrapped_lines = [
            line + "-" if i < len(wrapped_lines) - 1 else line
            for i, line in enumerate(wrapped_lines)
        ]
    return "\n".join(wrapped_lines)


def create_agent_interactive():
    name = click.prompt("Name")
    input_type = click.prompt(
        "Load an instructions file or enter instructions manually?",
        type=click.Choice(["f", "m"]),
        default="m",
    )
    if input_type == "m":
        instructions = click.prompt("Instructions")
    else:
        file_path = click.prompt("Please enter a file path", type=FilePathType())
        click.echo(f"Loading filepath {file_path}")
        with open(file_path, "r") as filehandle:
            instructions = filehandle.read()
    new_assistant = create_assistant_wrapper(name=name, instructions=instructions)
    save_instructions(new_assistant, new_assistant.instructions)
    click.echo(f"created {new_assistant.id} ({new_assistant.name})")
    return new_assistant


def select_assistant():
    assistants = list_assistants()
    current_assistant = None
    if len(assistants) > 0:
        assistant_choice = int(click.prompt("Select Assistant"))
        current_assistant = assistants[assistant_choice]
    else:
        exit("must create assistant")
    return current_assistant


def update_agent():
    """Update an agents instructions or file list"""
    assistant = select_assistant()
    current_assistant = assistant
    current_assistant_id = assistant.id
    formatted_instructions = current_assistant.instructions  # wrap_text()
    click.echo(f"current instructions:\n\n\t{formatted_instructions}\n")
    click.echo(f"current files:\n{current_assistant.file_ids}\n")
    if click.confirm(f"Update prompt?"):
        if click.confirm(f"did you update {current_assistant.id}/instructions.txt?"):
            new_instructions = None
        else:
            file_path = click.prompt("Please enter a file path", type=FilePathType())
            click.echo(f"Loading filepath {file_path}")
            with open(file_path, "r") as filehandle:
                new_instructions = filehandle.read()
        update_agent_instructions(
            current_assistant_id, new_instructions=new_instructions
        )

    if click.confirm(f"Update agents files?"):
        # select file from list of files
        update_agent_with_file(current_assistant_id)


def select_thread():
    thread = None
    threads = list_threads()

    if threads and len(threads) == 0:
        click.echo("No current threads.")

    if click.confirm("Create new thread?"):
        thread_name = click.prompt("Thread name?")
        new_thread = create_thread(thread_name=thread_name)
        threads.append(new_thread)

    # choose thread to send message to
    for index, thread in enumerate(threads):
        click.echo(f"{index}. {thread[0]}")

    thread_choice = int(click.prompt("Choose a thread."))
    current_thread = threads[thread_choice]
    return current_thread


def choose_or_create_file():
    if click.confirm("Create new file?"):
        file_path = file_path = click.prompt(
            "Please enter a file path", type=FilePathType()
        )
        if click.confirm(f"Are you sure you want to upload {file_path}?"):
            file = upload_file(file_path)
            file_id = file.id
    else:
        file_id = select_file_id()
    return file_id


def update_agent_with_file(assistant_id):
    # get agent
    my_assistant = client.beta.assistants.retrieve(assistant_id)
    chosen_file_id = choose_or_create_file()
    my_assistant.file_ids.append(chosen_file_id)
    # update agent
    client.beta.assistants.update(
        my_assistant.id,
        instructions=my_assistant.instructions,
        name=my_assistant.name,
        tools=my_assistant.tools,
        model=my_assistant.model,
        file_ids=my_assistant.file_ids,
    )


def run_thread(current_thread_id, current_assistant_id):
    # run job
    current_run = create_run(current_thread_id, current_assistant_id)
    current_run_id = current_run.id

    # wait for result
    result = wait_for_or_cancel_run(current_thread_id, current_run_id)
    if not result:
        return

    # when complete get thread/messages
    thread_messages = get_messages(current_thread_id)
    for content in thread_messages[0].content:
        save_local_message(thread_message=thread_messages[0], role="assistant")
        click.echo(content.text.value)


def update_agent_instructions(assistant_id, new_instructions=None):
    # get agent
    my_assistant = client.beta.assistants.retrieve(assistant_id)

    if new_instructions:
        # save them
        save_instructions(my_assistant, new_instructions)
    else:
        # they are updated locally
        new_instructions = load_instructions(my_assistant)

    if click.confirm(f"new instructions OK? \n{new_instructions}"):
        # change instructions
        my_assistant.instructions = new_instructions
        # update agent
        client.beta.assistants.update(
            my_assistant.id,
            instructions=my_assistant.instructions,
            name=my_assistant.name,
            tools=my_assistant.tools,
            model=my_assistant.model,
            file_ids=my_assistant.file_ids,
        )


def update_agent_tools(assistant_id, tools_to_add=None, tools_to_remove=None):
    pass


def select_file_id():
    all_files = list_files()
    choices = [
        (file.filename, file.id, file.created_at)
        for file in all_files
        if file.purpose == "assistants"
    ]
    indexes = []
    for index, (filename, fileid, created) in enumerate(choices):
        indexes.append(str(index))
        dt = datetime.fromtimestamp(created)
        created = dt.strftime("%Y-%m-%d-%H:%M:%S")
        click.echo(f"{index}. {filename} ({created})")

    index_choice = click.prompt("Choose a file index", type=click.Choice(indexes))

    file_choice = choices[int(index_choice)]
    click.echo(f"You chose file {file_choice[0]} ({file_choice[1]})")
    return file_choice[1]


def list_assistants():
    assistants = []
    click.echo("Listing assistants...")
    assistant_count = 0
    for assistant in get_assistants():
        assistants.append(assistant)
        click.echo(f"{assistant_count}. {assistant.name}({assistant.id})")
        assistant_count += 1

    if len(assistants) == 0:
        click.echo("No assistants not found.")
    return assistants



import click
from .oai import (
    # ALL_TOOLS_AVAILABLE,
    create_message,
    download_all_files,
    list_all_files,
    upload_file,
    delete_file,
    delete_assistant,
)
from .oai_wrappers import (
    create_agent_interactive,
    list_assistants,
    select_assistant,
    select_thread,
    update_agent,
    choose_or_create_file,
    run_thread,
    select_file_id,
)
from . import ASCII_ART, FilePathParamType


@click.group()
def cli():
    """Main entry point for oaicli."""
    pass


@cli.command(name="start")
def start_up():
    """Get Started"""
    click.echo(ASCII_ART)
    assistants = []
    current_assistant = None
    current_assistant_id = None

    if click.confirm("Create new assistant?"):
        new_assistant = create_agent_interactive()
        assistants.append(new_assistant)
        current_assistant = new_assistant
        current_assistant_id = new_assistant.id
    else:
        assistant = select_assistant()
        current_assistant = assistant
        current_assistant_id = assistant.id

    current_thread = select_thread()
    current_thread_id = current_thread[1]
    current_thread_name = current_thread[0]

    # thread loop
    click.echo("Ready.")
    click.echo(
        """
Inline commands:
 - Change agent by typing "change" (changes which agent runs the thread)
 - Add file to agent (or change prompt) with "agent"
 - Type 'exit' when done."""
    )
    while True:
        user_query = click.prompt(f"oaicli ({current_thread_name}) >")
        if user_query.strip() == "exit":
            exit("bye")
        elif user_query.strip() == "change":
            click.echo("changing agent")
            continue
        elif user_query.strip() == "agent":
            update_agent()

        if click.confirm("Add a file to the message?"):
            file_id = choose_or_create_file()
            thread_message = create_message(
                message_content=user_query,
                thread_name=current_thread[1],
                thread_id=current_thread_id,
                file_ids=[file_id],
            )
        else:
            # add message to thread
            thread_message = create_message(
                message_content=user_query,
                thread_name=current_thread[1],
                thread_id=current_thread_id,
            )

        click.echo(f"sent message {thread_message.id}")
        run_thread(current_thread_id, current_assistant_id)


@click.group()
def file():
    """Subcommand for managing file."""
    pass


cli.add_command(file)


@file.command(name="download-all")
def download_all():
    """Download all files"""
    download_all_files()


@file.command(name="list")
def list_all():
    """List all files"""
    list_all_files()


@file.command(name="upload")
@click.option(
    "-f", "--file-path", type=FilePathParamType(), help="Path to a file.", required=True
)
def do_upload_file(file_path):
    """Upload file"""
    if click.confirm(f"Are you sure you want to upload {file_path}?"):
        upload_file(file_path)


@file.command(name="delete")
def do_delete_file():
    """delete file"""
    file_id = select_file_id()
    if click.confirm(f"Are you sure you want to delete file {file_id}?"):
        delete_file(file_id)


# @file.command(name="list_all")
# def upload_directory():
#     """List all files"""
#     list_all_files()


@click.group()
def agent():
    """Subcommand for managing agents."""
    pass


cli.add_command(agent)


@agent.command(name="list")
def list_agents():
    """List agents."""
    list_assistants()


@agent.command(name="create")
def create_agent():
    """Create a new agent."""
    create_agent_interactive()


@agent.command(name="update")
def do_update_agent():
    """Update an agents instructions or file list"""
    update_agent()


@agent.command(name="delete")
def do_delete_agent():
    """delete an agent"""
    assistant = select_assistant()
    if click.confirm(
        f"Are you sure you want to delete assistant {assistant.name} ({assistant.id})?"
    ):
        delete_assistant(assistant.id)


# @agent.command(name="delete")
# @click.argument("agent")
# def delete_agent(agent):
#     """Delete an agent."""
#     click.echo(f"Deleting agent: {agent}")


# @agent.command(name="file-upload")
# @click.argument("agent")
# @click.argument("path")
# def file_upload(agent, path):
#     """Send a file to your agent."""
#     # creates a file

#     # get agent

#     # update agent

#     # adds file to agent
#     click.echo(f"Uploading file from {path} to agent {agent}")


# @agent.command(name="directory-upload")
# @click.argument("agent")
# @click.argument("path")
# def directory_upload(agent, path):
#     """Uploads all text and PDF files from a directory."""
#     click.echo(f"Uploading files from directory {path} to agent {agent}")


# @agent.command(name="url-upload")
# @click.argument("agent")
# @click.argument("url")
# def url_upload(agent, url):
#     """Upload content from a URL to an agent."""
#     click.echo(f"Uploading content from {url} to agent {agent}")


# @agent.command(name="file-list")
# @click.argument("agent")
# def file_list(agent):
#     """List an agent's files."""
#     click.echo(f"Listing files for agent {agent}")


def main():
    cli(auto_envvar_prefix="OAICLI")


if __name__ == "__main__":
    main()



import os
import appdirs
from dotenv import load_dotenv
from pathlib import Path
import click

current_working_directory = Path.cwd()
env_path = os.path.join(current_working_directory, ".env")
load_dotenv(dotenv_path=env_path)

OAICLI_HOME = ".oaicli"

home_dir = appdirs.user_config_dir(f"{current_working_directory}/{OAICLI_HOME}")
agents_dir = os.path.join(home_dir, "agents")
os.makedirs(agents_dir, exist_ok=True)
threads_dir = os.path.join(home_dir, "threads")
os.makedirs(threads_dir, exist_ok=True)
files_dir = os.path.join(home_dir, "files")
os.makedirs(files_dir, exist_ok=True)


DEFAULT_MODEL = "gpt-4-1106-preview"
DEFAULT_IMAGE_MODEL = "gpt-4-1106-vision-preview"

# Now you can access the API key
OPENAI_API_KEY = os.getenv("OPENAI_API_KEY")
OPEN_AI_MODEL_TYPE = os.getenv("OPEN_AI_MODEL_TYPE", DEFAULT_MODEL)
OPEN_AI_VISION_MODEL_TYPE = os.getenv("OPEN_AI_VISION_MODEL_TYPE", DEFAULT_IMAGE_MODEL)


# use with @click.argument('file_path', type=FilePathParamType())
class FilePathParamType(click.Path):
    def shell_complete(self, ctx, param, incomplete):
        return [
            click.shell_completion.Candidate(e)
            for e in os.listdir(".")
            if e.startswith(incomplete)
        ]


# use with prompt click.prompt('Please enter a file path', type=FilePathType())
class FilePathType(click.Path):
    def convert(self, value, param, ctx):
        if not os.path.exists(value):
            self.fail(f"The file path '{value}' does not exist.", param, ctx)
        return super().convert(value, param, ctx)


ASCII_ART = """

      ___           ___                       ___
     /  /\\         /  /\\        ___          /  /\\                      ___
    /  /::\\       /  /::\\      /  /\\        /  /:/                     /  /\\
   /  /:/\\:\\     /  /:/\\:\\    /  /:/       /  /:/       ___     ___   /  /:/
  /  /:/  \\:\\   /  /:/~/::\\  /__/::\\      /  /:/  ___  /__/\\   /  /\\ /__/::\\
 /__/:/ \\__\\:\\ /__/:/ /:/\\:\\ \\__\\/\\:\\__  /__/:/  /  /\\ \\  \\:\\ /  /:/ \\__\\/\\:\\__
 \\  \\:\\ /  /:/ \\  \\:\\/:/__\\/    \\  \\:\\/\\ \\  \\:\\ /  /:/  \\  \\:\\  /:/     \\  \\:\\/\\
  \\  \\:\\  /:/   \\  \\::/          \\__\\::/  \\  \\:\\  /:/    \\  \\:\\/:/       \\__\\::/
   \\  \\:\\/:/     \\  \\:\\          /__/:/    \\  \\:\\/:/      \\  \\::/        /__/:/
    \\  \\::/       \\  \\:\\         \\__\\/      \\  \\::/        \\__\\/         \\__\\/
     \\__\\/         \\__\\/                     \\__\\/
"""



from . import (
    OPEN_AI_MODEL_TYPE,
    DEFAULT_IMAGE_MODEL,
    agents_dir,
    files_dir,
    threads_dir,
)
from openai import OpenAI
import os
import logging
import click
import time
from typing import Optional, List
from pathlib import Path
import shutil
from datetime import datetime

JSON_RESPONSE_TYPE = {"type": "json_object"}
TEXT_RESPONSE_TYPE = {"type": "text"}
ALL_TOOLS_AVAILABLE = []
MAX_AGENT_LIST = 100
thread_separator = "---"
agent_separator = "---"
filename_separator = "---"
DEFAULT_TOOLS = [{"type": "retrieval"}]
logger = logging.getLogger("oaicli")

client = OpenAI()


def _get_assistant_path(my_assistant):
    return f"{agents_dir}/{my_assistant.id}"


def get_assistants(after=None):
    assistants = client.beta.assistants.list(
        order="desc",
        limit=MAX_AGENT_LIST,
        after=after,
    )
    for assistant in assistants:
        yield assistant
    if assistants.has_more:
        click.echo(f"There are more, but we stop at {MAX_AGENT_LIST}")


def create_assistant_wrapper(
    name: str, instructions: str, tools: Optional[List[str]] = None
):
    my_assistant = client.beta.assistants.create(
        instructions=instructions,
        name=name,
        tools=DEFAULT_TOOLS,
        model=OPEN_AI_MODEL_TYPE,
    )

    os.makedirs(_get_assistant_path(my_assistant), exist_ok=True)

    return my_assistant


def save_instructions(my_assistant, content):
    os.makedirs(_get_assistant_path(my_assistant), exist_ok=True)
    filepath = f"{_get_assistant_path(my_assistant)}/instructions.txt"
    file_object = open(filepath, "w")
    file_object.write(content)


def load_instructions(my_assistant):
    os.makedirs(_get_assistant_path(my_assistant), exist_ok=True)
    filepath = f"{_get_assistant_path(my_assistant)}/instructions.txt"
    file_object = open(filepath, "r")
    return file_object.read()


def list_threads():
    # seems theres no api call to list threads? can't be named?
    for first_level in os.walk(threads_dir):
        thread_dir = first_level[1]
        break
    return [thread.split(thread_separator) for thread in thread_dir]


def _get_thread_directory(thread):
    return f"{threads_dir}/{thread.metadata['name']}{thread_separator}{thread.id}"


def create_thread(thread_name: str):
    metadata = {"name": thread_name}
    empty_thread = client.beta.threads.create(metadata=metadata)
    thread_id = empty_thread.id
    click.echo(f"created thread {thread_id}.")
    os.makedirs(_get_thread_directory(empty_thread), exist_ok=True)
    return (thread_name, thread_id)


def save_local_message(thread_message: str, role: str):
    # TODO save content locally in thread directory
    thread = client.beta.threads.retrieve(thread_message.thread_id)
    thread_directory = _get_thread_directory(thread)
    now = datetime.now()
    timestamp = now.strftime("%Y-%m-%d-%H:%M:%S")
    local_path = f"{thread_directory}/{timestamp}-{role}-{thread_message.id}.txt"
    file_object = open(local_path, "w")
    all_content = ""
    for _content in thread_message.content:
        all_content += _content.text.value
    file_object.write(all_content)


def create_message(message_content: str, thread_name: str, thread_id: str):
    role = "user"
    thread_message = client.beta.threads.messages.create(
        thread_id,
        role=role,
        content=message_content,
    )
    save_local_message(thread_message, role=role)
    return thread_message


def create_run(thread_id, assistant_id):
    return client.beta.threads.runs.create(
        thread_id=thread_id, assistant_id=assistant_id
    )


def get_messages(thread_id):
    thread_messages = client.beta.threads.messages.list(thread_id)
    return thread_messages.data


def wait_for_or_cancel_run(thread_id, run_id):
    timeout_is_ok = True
    total_time = 0
    MAX_TIME = 46
    CHECK_INCREMENT = 2
    click.echo(f"Running for a max of {MAX_TIME} seconds.")
    while timeout_is_ok:
        run = client.beta.threads.runs.retrieve(thread_id=thread_id, run_id=run_id)
        if run.status != "completed":
            click.echo(
                f"Job is {run.status}. {total_time} s passed.", nl=False
            )
            time.sleep(CHECK_INCREMENT)
            click.echo("\r", nl=False)
            total_time += CHECK_INCREMENT
            if run.status in ["cancelled", "failed", "expired"]:
                return
            if total_time >= MAX_TIME:
                run = client.beta.threads.runs.cancel(
                    thread_id="thread_abc123", run_id="run_abc123"
                )
        elif run.status == "completed":
            return True


def _get_local_filepath(file):
    new_filename = f"{file.id}{filename_separator}{file.filename}"
    return f"{files_dir}/{new_filename}"


def delete_file(file_id: str):
    client.files.delete(file_id)


def delete_assistant(assistant_id):
    client.beta.assistants.delete(assistant_id)


def upload_file(local_filepath: str):
    file = client.files.create(file=open(local_filepath, "rb"), purpose="assistants")
    file_path = Path(local_filepath)
    filename = file_path.name

    if filename != file.filename:
        click.echo(f"warning {filename} != file.filename")

    shutil.copyfile(local_filepath, _get_local_filepath(file))
    return file


def list_files():
    return client.files.list()


def download_all_files():
    # TODO -  option for file ids in use, not in use
    all_files = list_files()
    for file in all_files:
        local_path = _get_local_filepath(file)
        if not os.path.exists(local_path):
            content = client.files.retrieve_content(file.id)
            file_object = open(local_path, "w")
            file_object.write(content)
            click.echo(f"{file.filename}\t\t({file.id}) to {local_path}")
        else:
            click.echo(f"{file.filename}\t\t({file.id}) exists locally")


def list_all_files():
    # TODO -  option for file ids in use, not in use
    all_files = list_files()
    for file in all_files:
        local_path = _get_local_filepath(file)
        dt = datetime.fromtimestamp(file.created_at)
        created = dt.strftime("%Y-%m-%d-%H:%M:%S")
        file_description = f"{file.filename}\t\t({created} {file.purpose} {file.id})"
        if not os.path.exists(local_path):
            click.echo(f"{file_description}) does not exist locally")
        else:
            click.echo(f"{file_description}) exists locally")


# def vision_url(prompt: str, image_url: str):
#     response = client.chat.completions.create(
#         model=DEFAULT_IMAGE_MODEL,
#         messages=[
#             {
#                 "role": "user",
#                 "content": [
#                     {"type": "text", "text": prompt},
#                     {
#                         "type": "image_url",
#                         "image_url": {
#                             "url": image_url,
#                             "detail": "low",
#                         },
#                     },
#                 ],
#             }
#         ],
#         max_tokens=300,
#     )

#     return response
